from enum import Enum
from typing import Union, Type

from PIL import Image

from ..util.type import E


class ResizeMethod(Enum):
    """Resize methods can be used in :func:`~resize_image` and :func:`~resize_video`"""
    BICUBIC = Image.BICUBIC
    BILINEAR = Image.BILINEAR
    NEAREST = Image.NEAREST
    LANCZOS = Image.LANCZOS
    BOX = Image.BOX


class PaddingMethod(Enum):
    """The padding methods used in :func:`~padding_video`"""
    ZERO = "zero"
    SAME = "same"


class PaddingPosition(Enum):
    """The padding position used in :func:`~padding_video`"""
    HEAD = "head"
    TAIL = "tail"
    AVERAGE = "average"


def get_enum_value(enum: Union[E, str], EnumType: Type[E]) -> str:
    if type(enum) == EnumType:
        return enum.value
    elif type(enum) == str:
        return enum.lower()
    else:
        raise TypeError("Not matched enum type.")



from .iou import iou_1d, iou_2d
from .psnr import psnr_video, psnr_image
from .ssim import ssim_video, ssim_image
from .secs import secs
from .fid import FID

__all__ = [
    "iou_1d",
    "iou_2d",
    "psnr_video",
    "psnr_image",
    "ssim_video",
    "ssim_image",
    "secs",
    "FID",
]



from numpy import ndarray
from torch import Tensor

from tensorneko_util.util import dispatch, Eval

from tensorneko_util.io import read


@Eval.later
def _secs_encoder():
    from resemblyzer import VoiceEncoder
    return VoiceEncoder()


@dispatch
def secs(pred: str, real: str) -> float:
    from resemblyzer import VoiceEncoder, preprocess_wav
    pred_audio = preprocess_wav(read.audio(pred).audio[0].numpy())
    real_audio = preprocess_wav(read.audio(real).audio[0].numpy())
    return _secs_compute(pred_audio, real_audio)


@dispatch
def secs(pred: Tensor, real: Tensor) -> float:
    return secs(pred.numpy(), real.numpy())


@dispatch
def secs(pred: ndarray, real: ndarray) -> float:
    from resemblyzer import VoiceEncoder, preprocess_wav
    if len(pred.shape) == 2:
        if pred.shape[0] == 1:
            pred = pred.squeeze(0)
        elif pred.shape[1] == 1:
            pred = pred.squeeze(1)
        else:
            raise ValueError("The input audio must be mono.")

    if len(real.shape) == 2:
        if real.shape[0] == 1:
            real = real.squeeze(0)
        elif real.shape[1] == 1:
            real = real.squeeze(1)
        else:
            raise ValueError("The input audio must be mono.")

    pred_audio = preprocess_wav(pred)
    real_audio = preprocess_wav(real)

    return _secs_compute(pred_audio, real_audio)


def _secs_compute(pred_audio: ndarray, real_audio: ndarray) -> float:
    encoder = _secs_encoder.value
    real_embed = encoder.embed_utterance(real_audio)
    pred_embed = encoder.embed_utterance(pred_audio)

    return float((real_embed * pred_embed).sum())



from abc import abstractmethod, ABC
from typing import Tuple

import numpy as np
from torch.utils.data.dataset import Dataset, T_co


class NestedDataset(Dataset[T_co], ABC):
    """
    A dataset that wraps multiple datasets.

    This class is a subclass of the PyTorch Dataset class and is used to wrap multiple datasets.
    It provides the necessary methods for a PyTorch Dataset, such as `__getitem__` and `load_data`.

    Args:
        outer_size (``int``): The number of datasets to be wrapped.

    Attributes:
        datasets (``List[Dataset[T_co]]``): The list of datasets that the NestedDataset wraps.
    """

    def __init__(self, outer_size: int):
        super().__init__()
        self.datasets = []
        for i in range(outer_size):
            self.datasets.append(self.load_data(i))

        lengths = np.array([len(dataset) for dataset in self.datasets])
        self._acc = np.cumsum(lengths)

    @abstractmethod
    def load_data(self, index: int) -> Dataset[T_co]:
        """
        Abstract method to load data for a specific index.

        Args:
            index (``int``): The index of the dataset to load.

        Returns:
            The loaded dataset.
        """
        ...

    def _query_index(self, index: int) -> Tuple[int, int]:
        """
        Retrieves the outer and inner indices for a specific index.

        Args:
            index (``int``): The index to query.

        Returns:
            A tuple containing the outer and inner indices.
        """
        outer_index = np.searchsorted(self._acc, index, "left")
        inner_index = index - self._acc[outer_index - 1]
        return outer_index, inner_index

    def __getitem__(self, index: int) -> T_co:
        """
        Retrieves an item from the nested dataset at a specific index.

        Args:
            index (``int``): The index of the item to retrieve.

        Returns:
            The item at the specified index.
        """
        outer_index, inner_index = self._query_index(index)
        return self.datasets[outer_index][inner_index]



from tensorneko_util.preprocess.face_detector import AbstractFaceDetector, OpencvFaceDetector, FaceXZooFaceDetector, \
    AnimeFaceDetector

__all__ = [
    "AbstractFaceDetector",
    "OpencvFaceDetector",
    "FaceXZooFaceDetector",
    "AnimeFaceDetector"
]



from typing import Sized

from torch.utils.data.sampler import Sampler, T_co


class SequentialIterSampler(Sampler[T_co]):
    """
    Use to split the large scale data into small subsets for each epochs
    For example, if the dataset size is 1M, and the num_samples = 1000, then each epoch will only use 1000 samples, and
    the next epoch will use the next 1000 samples.

    Args:
        data_source (:class:`torch.utils.data.Dataset`): The dataset to sample from.
        num_samples (``int``): The number of samples to draw from the dataset.

    Examples::

        dataset = torch.utils.data.TensorDataset(torch.randn(1000, 10))
        sampler = SequentialIterSampler(dataset, 100)
        for indexes in sampler:
            print(indexes)
        # [0, 1, 2, ..., 99]
        # [100, 101, 102, ..., 199]
        # [200, 201, 202, ..., 299]
    """

    def __init__(self, data_source: Sized, num_samples: int):
        super().__init__(data_source)
        self.data_source = data_source
        self.num_samples = num_samples
        self.total_size = len(data_source)
        self.current_position = 0

    def __iter__(self):
        yield from map(lambda x: x % self.total_size,
            range(self.current_position, self.current_position + self.num_samples))
        self.current_position = (self.current_position + self.num_samples) % self.total_size

    def __len__(self):
        return self.num_samples



from abc import ABC, abstractmethod
from typing import Optional, Union, Sequence, Dict

from torch import Tensor
from torch.nn import Module, MSELoss
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau

from ..neko_model import NekoModel


class AutoEncoder(NekoModel, ABC):

    def __init__(self, name: str, learning_rate: float, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.learning_rate = learning_rate
        self.encoder = self.build_encoder()
        self.decoder = self.build_decoder()
        self.loss_fn = MSELoss()

    def forward(self, x: Tensor) -> Tensor:
        z = self.encode(x)
        return self.decode(z)

    def encode(self, x: Tensor) -> Tensor:
        return self.encoder(x)

    def decode(self, z: Tensor) -> Tensor:
        return self.decoder(z)

    @abstractmethod
    def build_encoder(self) -> Module:
        pass

    @abstractmethod
    def build_decoder(self) -> Module:
        pass

    def training_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        optimizer_idx: Optional[int] = None, hiddens: Optional[Tensor] = None
    ) -> Dict[str, Tensor]:
        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]
        x_hat = self(x)
        loss = self.loss_fn(x_hat, x)
        return {"loss": loss}

    def validation_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:
        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]
        x_hat = self(x)
        loss = self.loss_fn(x_hat, x)
        return {"loss": loss}

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.learning_rate)
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": ReduceLROnPlateau(optimizer, patience=5, factor=0.5, min_lr=1e-8),
                "monitor": "val_loss"
            }
        }


import torch
from torch import Tensor

from tensorneko_util.util import dispatch
from tensorneko_util.util.dispatched_misc import sparse2binary as _sparse2binary
from tensorneko_util.util.dispatched_misc import binary2sparse as _binary2sparse


@dispatch.base(_sparse2binary)
def sparse2binary(x: Tensor, length: int = None) -> Tensor:
    """
    Convert a sparse tensor to a binary tensor.

    Args:
        x (:class:`~torch.Tensor`): The input sparse tensor.
        length (``int``, optional): The length of the binary tensor.

    Returns:
        :class:`~torch.Tensor`: The binary tensor.
    """
    return torch.zeros(length or (x.max() + 1), dtype=torch.int).scatter_(0, x, 1)


@dispatch.base(_binary2sparse)
def binary2sparse(x: Tensor) -> Tensor:
    """
    Convert a binary tensor to a sparse tensor.

    Args:
        x (:class:`~torch.Tensor`): The input binary tensor.

    Returns:
        :class:`~torch.Tensor`: The sparse tensor.
    """
    return x.nonzero().squeeze(1)



from typing import Optional, Union, Sequence, Dict, Any

from torch import Tensor
from torch.nn import BCEWithLogitsLoss, Module
from torch.optim import Adam
from torchmetrics import Accuracy, F1Score, AUROC

from ..neko_model import NekoModel
from ..util import Shape


class BinaryClassifier(NekoModel):

    def __init__(self, name, model: Module, learning_rate: float = 1e-4, distributed: bool = False,
        input_shape: Optional[Shape] = None
    ):
        super().__init__(name, input_shape=input_shape, distributed=distributed)
        self.save_hyperparameters()
        self.model = model
        self.learning_rate = learning_rate
        self.loss_fn = BCEWithLogitsLoss()
        self.acc_fn = Accuracy(task="binary")
        self.f1_fn = F1Score(task="binary")
        self.auc_fn = AUROC(task="binary")

    @classmethod
    def from_module(cls, model: Module, learning_rate: float = 1e-4, name: str = "binary_classifier",
        distributed: bool = False, input_shape: Optional[Shape] = None
    ):
        return cls(name, model, learning_rate, distributed, input_shape)

    def forward(self, x):
        return self.model(x)

    def step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]]) -> Dict[str, Tensor]:
        x, y = batch
        y_hat = self(x).squeeze(1)
        loss = self.loss_fn(y_hat, y)
        prob = y_hat.sigmoid()
        acc = self.acc_fn(prob, y)
        f1 = self.f1_fn(prob, y)
        auc = self.auc_fn(prob, y)
        return {"loss": loss, "metric/acc": acc, "metric/f1": f1, "metric/auc": auc}

    def training_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        optimizer_idx: Optional[int] = None, hiddens: Optional[Tensor] = None
    ) -> Dict[str, Tensor]:
        return self.step(batch)

    def validation_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:
        return self.step(batch)

    def predict_step(self, batch: Tensor, batch_idx: int, dataloader_idx: Optional[int] = None) -> Any:
        x, y = batch
        return self(x)

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.learning_rate)
        return [optimizer]



from lightning.pytorch import Callback, Trainer, LightningModule


class EpochNumLogger(Callback):
    """Log epoch number in each epoch start."""

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        key = "epoch/epoch"
        value = trainer.current_epoch
        pl_module.logger.log_metrics({key: value}, step=trainer.global_step)
        pl_module.log(key, value, logger=False, sync_dist=pl_module.distributed)



from typing import Union, Optional

from einops.layers.torch import Rearrange
from torch import Tensor
from torch.nn import Conv2d, Conv3d

from ..neko_module import NekoModule
from ..util import ModuleFactory, Shape, F


class Patching(NekoModule):
    """
    Patching input tensor with specified patch size.

    If the `patch_size` is tuple of dims, it will patch last same number of dims. E.g. if `patch_size=(16, 16)`, and the
        input tensor has shape `(b, c, h, w)`, it will patch `(b, c, h // 16, w // 16, 16, 16)`.

    If the `patch_size` is int, it will patch from dim 2 to last. E.g. if `patch_size=16`, and the input tensor has
        shape `(a, b, c, d, e, f)`, it will patch `(a, b, c // 16, d // 16, e // 16, f // 16)`.

    Args:
        patch_size (``int`` | :class:`~tensorneko.util.type.Shape`): The patch size.

    Examples::

        import torch
        x = torch.rand(2, 3, 224, 224)
        patching = Patching(patch_size=16)

        print(patching(x).shape)  # (2, 3, 14, 14, 16, 16)

    """

    def __init__(self, patch_size: Union[int, Shape]):
        super().__init__()
        self.patch_size = patch_size

    def forward(self, x: Tensor) -> Tensor:
        x_dim = x.dim()

        assert x_dim > 2, "The input must be at least 3-dimensional."
        if type(self.patch_size) is int:
            patch_dims = range(2, x_dim)
            self.patch_size = (self.patch_size,) * len(patch_dims)
        else:
            patch_dims = range(x_dim - len(self.patch_size), x_dim)

        for i, patch_dim in enumerate(patch_dims):
            x = x.unfold(patch_dim, self.patch_size[i], self.patch_size[i])

        return x


class PatchEmbedding2d(NekoModule):
    """
    The patch embedding layer from Vision Transformer (ViT) (Dosovitskiy, et al., 2020). This layer will take an image
    as input, and divide to several patches, and project to a provided number of channels.

    Here it is implemented by an equivalent convolution layer with same stride length and kernel size.

    Args:
        input_size (:class:`~tensorneko.util.type.Shape`): The size of input shape. (C, H, W).

        patch_size (``int`` | :class:`~tensorneko.util.type.Shape`): The side length of patch or a patch Shape of
            (H, W).

        embedding (``int``): Output embedding dimension.

        strides (``int`` | :class:`~tensorneko.util.type.Shape`, optional): The strides of the patch window.
            Default is the size of patch.

        build_normalization (``() -> torch.nn.Module``, optional): An normalization module builder to be used in the
            layer. Default ``None``.

    Examples::

        patch_emb = tensorneko.layer.PatchEmbedding2d(
            input_size=(3, 224, 224),
            patch_size=(16, 16),
            embedding=768
        )

        print(patch_emb(rand(8, 3, 224, 224)).shape)  # (8, 196, 768)

    References:
        Dosovitskiy, A., Beyer, L., Kolesnikov, A., Weissenborn, D., Zhai, X., Unterthiner, T., ... & Houlsby, N.
        (2020). An image is worth 16x16 words: Transformers for image recognition at scale.
        arXiv preprint arXiv:2010.11929.

    """

    def __init__(self, input_size: Shape, patch_size: Union[int, Shape], embedding: int,
        strides: Optional[Union[int, Shape]] = None,
        build_normalization: Optional[ModuleFactory] = None
    ):
        super().__init__()
        # channel, height, width
        c, h, w = input_size
        # patch_height, patch_width
        ph, pw = (patch_size, patch_size) if type(patch_size) is int else patch_size

        # configure the strides for conv2d
        if strides is None:
            # no specified means no overlap and gap between patches
            strides = (ph, pw)
        elif type(strides) is int:
            # transform the side length of strides to 2D
            strides = (strides, strides)

        self.projection = Conv2d(c, embedding, kernel_size=(ph, pw), stride=strides)
        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
        self.rearrange = Rearrange("b d nh nw -> b (nh nw) d")

    def forward(self, x: Tensor) -> Tensor:
        x = self.projection(x)
        x = self.rearrange(x)
        if self.has_norm:
            x = self.normalization(x)
        return x


class PatchEmbedding3d(NekoModule):
    """
    The 3D cube version of Patch Embedding layer.

    Args:
        input_size (:class:`~tensorneko.util.type.Shape`): The size of input shape. (C, T, H, W).
        patch_size (``int`` | :class:`~tensorneko.util.type.Shape`): The side length of patch or a patch Shape of
            (T, H, W).
        embedding (``int``): Output embedding dimension.
        strides (``int`` | :class:`~tensorneko.util.type.Shape`, optional): The strides of the patch window.
            Default is the size of patch.
        build_normalization (``() -> torch.nn.Module``, optional): An normalization module builder to be used in the
            layer. Default ``None``.

    Examples::

        patch_emb = tensorneko.layer.PatchEmbedding3d(
            input_size=(3, 16, 224, 224),
            patch_size=(2, 16, 16),
            embedding=1024
        )

        print(patch_emb(rand(8, 3, 16, 224, 224)).shape)  # (8, 1568, 1024)

    """

    def __init__(self, input_size: Shape, patch_size: Union[int, Shape], embedding: int,
        strides: Optional[Union[int, Shape]] = None,
        build_normalization: Optional[ModuleFactory] = None
    ):
        super().__init__()
        # channel, time, height, width
        c, t, h, w = input_size
        # patch_time, patch_height, patch_width
        pt, ph, pw = (patch_size, patch_size, patch_size) if type(patch_size) is int else patch_size

        # configure the strides for conv3d
        if strides is None:
            # no specified means no overlap and gap between patches
            strides = (pt, ph, pw)
        elif type(strides) is int:
            # transform the side length of strides to 3D
            strides = (strides, strides, strides)

        self.projection = Conv3d(c, embedding, kernel_size=(pt, ph, pw), stride=strides)
        self.has_norm = build_normalization is not None
        if self.has_norm:
            self.normalization = build_normalization()
        self.rearrange = Rearrange("b d nt nh nw -> b (nt nh nw) d")

    def forward(self, x: Tensor) -> Tensor:
        x = self.projection(x)
        x = self.rearrange(x)
        if self.has_norm:
            x = self.normalization(x)
        return x



from typing import Any

from lightning.pytorch import Callback, Trainer, LightningModule
from lightning.pytorch.utilities.types import STEP_OUTPUT


class GpuStatsLogger(Callback):
    """Log GPU stats for each training epoch"""

    def __init__(self, delay: float = 0.5, on_epoch: bool = True, on_step: bool = False):
        try:
            from gpumonitor.monitor import GPUStatMonitor
        except ImportError:
            raise ImportError("gpumonitor is required to use GPUStatsLogger")

        self.monitor_epoch = GPUStatMonitor(delay=delay) if on_epoch else None
        self.monitor_step = GPUStatMonitor(delay=delay) if on_step else None
        self.on_epoch = on_epoch
        self.on_step = on_step
        assert self.on_epoch or self.on_step, "on_epoch and on_step cannot be both False"

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if not self.on_epoch:
            return
        self.monitor_epoch.reset()

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if not self.on_epoch:
            return
        for gpu in self.monitor_epoch.average_stats.gpus:
            logged_info = {
                f"system/gpu{gpu.index}_memory_used_epoch": gpu.memory_used / 1024,
                f"system/gpu{gpu.index}_memory_total_epoch": gpu.memory_total / 1024,
                f"system/gpu{gpu.index}_memory_util_epoch": gpu.memory_used / gpu.memory_total,
                f"system/gpu{gpu.index}_temperature_epoch": float(gpu.temperature),
                f"system/gpu{gpu.index}_utilization_epoch": gpu.utilization / 100,
                f"system/gpu{gpu.index}_power_draw_epoch": float(gpu.power_draw),
                f"system/gpu{gpu.index}_power_percentage_epoch": gpu.power_draw / gpu.power_limit,
                f"system/gpu{gpu.index}_fan_speed_epoch": float(gpu.fan_speed) if gpu.fan_speed is not None else 0.,
            }
            pl_module.logger.log_metrics(logged_info, step=trainer.global_step)
            pl_module.log_dict(logged_info, logger=False, sync_dist=pl_module.distributed)

    def on_train_batch_start(
        self, trainer: Trainer, pl_module: LightningModule, batch: Any, batch_idx: int
    ) -> None:
        if not self.on_step:
            return
        self.monitor_step.reset()

    def on_train_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        if not self.on_step:
            return
        for gpu in self.monitor_step.average_stats.gpus:
            logged_info = {
                f"system/gpu{gpu.index}_memory_used_step": gpu.memory_used / 1024,
                f"system/gpu{gpu.index}_memory_total_step": gpu.memory_total / 1024,
                f"system/gpu{gpu.index}_memory_util_step": gpu.memory_used / gpu.memory_total,
                f"system/gpu{gpu.index}_temperature_step": float(gpu.temperature),
                f"system/gpu{gpu.index}_utilization_step": gpu.utilization / 100,
                f"system/gpu{gpu.index}_power_draw_step": float(gpu.power_draw),
                f"system/gpu{gpu.index}_power_percentage_step": gpu.power_draw / gpu.power_limit,
                f"system/gpu{gpu.index}_fan_speed_step": float(gpu.fan_speed) if gpu.fan_speed is not None else 0.,
            }
            pl_module.logger.log_metrics(logged_info, step=trainer.global_step)
            pl_module.log_dict(logged_info, logger=False, sync_dist=pl_module.distributed)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union, List

import torch
import torchvision.io
from torch import Tensor
from torch.types import Device
from torch.utils.data import IterableDataset
from torchmetrics.image.fid import FrechetInceptionDistance
from torchvision.transforms.functional import resize

from tensorneko_util.backend import VisualLib
from tensorneko_util.backend._tqdm import import_tqdm_auto

try:
    from typing import Literal
    TypeOption = Literal["video", "image"]
except ImportError:
    TypeOption = str


class FID:
    """
    Calculate Fréchet inception distance based on torchmetrics. Require library "torch-fidelity".

    Args:
        device (``str`` | :class:`~torch.device`, optional): Device to run the metric. Default: ``"cpu"``.

    Example::

        from tensorneko.evaluation import FID
        fid = FID("cuda")
        
        # add predicted and real images
        fid.add_pred_image("path/to/pred/image1.png")
        fid.add_pred_image("path/to/pred/image2.png")
        fid.add_true_image("path/to/true/image1.png")
        fid.add_true_image("path/to/true/image2.png")

        # add predicted and real videos
        fid.add_pred_video("path/to/pred/video1.mp4")
        fid.add_pred_video("path/to/pred/video2.mp4")
        fid.add_true_video("path/to/true/video1.mp4")
        fid.add_true_video("path/to/true/video2.mp4")

        # compute FID
        fid_score = fid.compute(batch_size=128, num_workers=8, progress_bar=True)
        print(fid_score)

    """

    def __init__(self, device: Union[str, Device] = "cpu"):
        self.device = torch.device(device)
        self.fid = FrechetInceptionDistance().to(self.device)
        self.pred_data = _FIDDataset()
        self.true_data = _FIDDataset()

    def add_pred_image(self, path: str) -> FID:
        self.pred_data.add_image(path)
        return self

    def add_pred_video(self, path: str) -> FID:
        self.pred_data.add_video(path)
        return self

    def add_true_image(self, path: str) -> FID:
        self.true_data.add_image(path)
        return self

    def add_true_video(self, path: str) -> FID:
        self.true_data.add_video(path)
        return self

    def to(self, device: Union[str, Device]) -> FID:
        self.device = torch.device(device)
        self.fid.to(self.device)
        return self

    def cpu(self) -> FID:
        return self.to("cpu")

    def cuda(self) -> FID:
        return self.to("cuda")

    def compute(self, batch_size=128, num_workers=0, progress_bar: bool = False) -> float:
        pred = torch.utils.data.DataLoader(self.pred_data, batch_size=batch_size, num_workers=num_workers)
        true = torch.utils.data.DataLoader(self.true_data, batch_size=batch_size, num_workers=num_workers)

        if progress_bar:
            tqdm = import_tqdm_auto().tqdm
            pred = tqdm(pred, total=len(pred), desc="Forward predicted features")
            true = tqdm(true, total=len(true), desc="Forward ground truth features")

        for batch in pred:
            self.fid.update(batch.to(self.device), real=False)
        for batch in true:
            self.fid.update(batch.to(self.device), real=True)

        return self.fid.compute().item()

    def reset(self):
        self.pred_data = _FIDDataset()
        self.true_data = _FIDDataset()
        self.fid.reset()


@dataclass
class _FIDEntry:
    type: TypeOption
    path: str


class _FIDDataset(IterableDataset):

    def __init__(self):
        super().__init__()
        self.content: List[_FIDEntry] = []
        self.cv2 = None
        self.length = 0

    def add_image(self, path: str):
        self.content.append(_FIDEntry("image", path))
        self.length += 1

    def add_video(self, path: str):
        if self.cv2 is None:
            if not VisualLib.opencv_available():
                raise RuntimeError("OpenCV is required to add video. Please install it first.")
            else:
                import cv2
                self.cv2 = cv2
        self.content.append(_FIDEntry("video", path))
        cap = self.cv2.VideoCapture(path)
        if not cap.isOpened():
            raise RuntimeError("Cannot open video file.")
        n_frames = int(cap.get(self.cv2.CAP_PROP_FRAME_COUNT))
        self.length += n_frames
        cap.release()

    @staticmethod
    def _preprocess_image(image: Tensor) -> Tensor:
        image = resize(image, [299, 299])
        return image

    def _read_image(self, path: str) -> torch.Tensor:
        image = torchvision.io.read_image(path)
        image = self._preprocess_image(image)
        return image

    def _read_video(self, path: str) -> torch.Tensor:
        cap = self.cv2.VideoCapture(path)
        if not cap.isOpened():
            raise RuntimeError("Cannot open video file.")

        while cap.isOpened():
            ret, frame = cap.read()
            if not ret:
                break

            frame = self.cv2.cvtColor(frame, self.cv2.COLOR_BGR2RGB)
            frame = torch.from_numpy(frame).permute(2, 0, 1)
            frame = self._preprocess_image(frame)
            yield frame

        cap.release()

    def __iter__(self):
        for entry in self.content:
            if entry.type == "image":
                yield self._read_image(entry.path)
            elif entry.type == "video":
                yield from self._read_video(entry.path)
            else:
                raise RuntimeError("Unknown entry type.")

    def __len__(self):
        return self.length



import re
import subprocess

import torch
from torch import Tensor, tensor
from torchmetrics.functional import peak_signal_noise_ratio as psnr

from tensorneko_util.backend.visual_lib import VisualLib
from tensorneko_util.util import dispatch
from .enum import Reduction
from ..io import read
from ..preprocess import padding_video, PaddingMethod


@dispatch
def psnr_image(pred: str, real: str) -> Tensor:
    """
    Calculate PSNR of an image.

    Args:
        pred (``str``): Path to the predicted image.
        real (``str``): Path to the real image.

    Returns:
        :class:`~torch.Tensor`: The psnr of the image.
    """
    pred_image = read.image(pred).unsqueeze(0)
    real_image = read.image(real).unsqueeze(0)
    return psnr(pred_image, real_image, data_range=1.0, reduction="mean")


@dispatch
def psnr_image(pred: Tensor, real: Tensor, reduction: Reduction = Reduction.MEAN) -> Tensor:
    """
    Calculate PSNR of an image.

    Args:
        pred (:class:`~torch.Tensor`): Predicted images tensor. (B, C, H, W) or (C, H, W)
        real (:class:`~torch.Tensor`): Real images tensor. (B, C, H, W) or (C, H, W)
        reduction: (``str``, optional): Reduction method "mean", "sum" and "none". Default: ``"mean"``.

    Returns:
        :class:`~torch.Tensor`: The ssim of the images.
    """
    if pred.dim() == 3:
        pred = pred.unsqueeze(0)

    if real.dim() == 3:
        real = real.unsqueeze(0)

    assert pred.shape[0] == real.shape[0], "The number of images in pred and real must be equal."

    if pred.dtype in (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64):
        pred = pred.float() / 255.

    if real.dtype in (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64):
        real = real.float() / 255.

    reduction_method = reduction.value

    if reduction_method == "mean":
        reduction_method = "elementwise_mean"
        dim = None
    else:
        dim = (1, 2, 3)

    return psnr(pred, real, data_range=1.0, reduction=reduction_method, dim=dim)


@dispatch
def psnr_video(pred: str, real: str, use_ffmpeg: bool = False) -> Tensor:
    """
    Calculate PSNR of a video.

    Args:
        pred (``str``): Path to the predicted video.
        real (``str``): Path to the real video.
        use_ffmpeg (``bool``, optional): Whether to use ffmpeg to calculate psnr.

    Returns:
        :class:`~torch.Tensor`: The psnr of the video.
    """
    if use_ffmpeg:
        if not VisualLib.ffmpeg_available():
            raise RuntimeError("ffmpeg is not found.")

        p = subprocess.run(["ffmpeg", "-i", pred, "-i", real, "-filter_complex", "psnr", "-f", "null", "/dev/null"],
                           capture_output=True)
        psnr_out = p.stderr.decode().split("\n")[-2]
        return tensor(float(re.search(r" average:([\d.]+) ", psnr_out)[1]))
    else:
        pred_video = tensor(read.video(pred).video)  # (T, C, H, W)
        real_video = tensor(read.video(real).video)
        return psnr_video(pred_video, real_video)


@dispatch
def psnr_video(pred: Tensor, real: Tensor) -> Tensor:
    """
    Calculate PSNR of a video.

    Args:
        pred (:class:`~torch.Tensor`): Predicted video tensor. (T, C, H, W)
        real (:class:`~torch.Tensor`): Real video tensor. (T, C, H, W)

        Returns:
            :class:`~torch.Tensor`: The psnr of the video.
    """
    if pred.dtype in (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64):
        pred = pred.float() / 255.0
    if real.dtype in (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64):
        real = real.float() / 255.0

    real_video = padding_video(real, pred.shape[0], PaddingMethod.SAME)
    return psnr_image(pred, real_video, reduction=Reduction.MEAN)



import os
from datetime import timedelta
from time import time
from typing import Optional, Union, List, Dict

from lightning.fabric.plugins.precision.precision import _PRECISION_INPUT
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch import Trainer, Callback
from lightning.pytorch.accelerators import Accelerator
from lightning.pytorch.callbacks import ModelCheckpoint, Checkpoint
from lightning.pytorch.loggers import Logger, TensorBoardLogger
from lightning.pytorch.profilers import Profiler
from lightning.pytorch.strategies import Strategy
from lightning.pytorch.trainer.connectors.accelerator_connector import _LITERAL_WARN

try:
    from lightning.pytorch.plugins import PLUGIN_INPUT
except ImportError:
    from lightning.pytorch.plugins import _PLUGIN_INPUT as PLUGIN_INPUT

from .callback import NilCallback, LrLogger, EpochNumLogger, EpochTimeLogger, GpuStatsLogger, SystemStatsLogger


class NekoTrainer(Trainer):

    def __init__(self,
        logger: Optional[str],
        enable_checkpointing: bool = True,
        callbacks: Optional[Union[List[Callback], Callback]] = None,
        log_every_n_steps: int = 50,
        precision: _PRECISION_INPUT = "32-true",
        accelerator: Union[str, Accelerator] = "auto",
        strategy: Union[str, Strategy] = "auto",
        devices: Union[List[int], str, int] = "auto",
        num_nodes: int = 1,
        fast_dev_run: Union[int, bool] = False,
        max_epochs: Optional[int] = None,
        min_epochs: Optional[int] = None,
        max_steps: int = -1,
        min_steps: Optional[int] = None,
        max_time: Optional[Union[str, timedelta, Dict[str, int]]] = None,
        limit_train_batches: Optional[Union[int, float]] = None,
        limit_val_batches: Optional[Union[int, float]] = None,
        limit_test_batches: Optional[Union[int, float]] = None,
        limit_predict_batches: Optional[Union[int, float]] = None,
        overfit_batches: Union[int, float] = 0.0,
        val_check_interval: Optional[Union[int, float]] = None,
        check_val_every_n_epoch: Optional[int] = 1,
        num_sanity_val_steps: Optional[int] = None,
        enable_progress_bar: Optional[bool] = None,
        enable_model_summary: Optional[bool] = None,
        accumulate_grad_batches: int = 1,
        gradient_clip_val: Optional[Union[int, float]] = None,
        gradient_clip_algorithm: Optional[str] = None,
        deterministic: Optional[Union[bool, _LITERAL_WARN]] = None,
        benchmark: Optional[bool] = True,
        inference_mode: bool = True,
        use_distributed_sampler: bool = True,
        profiler: Optional[Union[Profiler, str]] = None,
        detect_anomaly: bool = False,
        barebones: bool = False,
        plugins: Optional[Union[PLUGIN_INPUT, List[PLUGIN_INPUT]]] = None,
        sync_batchnorm: bool = False,
        reload_dataloaders_every_n_epochs: int = 0,
        default_root_dir: Optional[_PATH] = None,
    ):
        # define a random logger name
        self.log_name = f"{logger}_{int(time())}"
        if callbacks is None:
            callbacks = []
        # build checkpoint callback or from user defined
        if enable_checkpointing and len([c for c in callbacks if isinstance(c, Checkpoint)]) == 0:
            # use default checkpoint callback
            new_callback = ModelCheckpoint(
                dirpath=os.path.join("logs", self.log_name, "checkpoints"),
                save_last=True, filename="{epoch}-{val_loss:.3f}", monitor="val_loss",
                mode="min"
            )
        elif enable_checkpointing and len([c for c in callbacks if isinstance(c, Checkpoint)]) > 0:
            # use user defined checkpoint callback
            new_callback = NilCallback()
        elif not enable_checkpointing and len([c for c in callbacks if isinstance(c, Checkpoint)]) > 0:
            # raise error because checkpoint callback is not allowed
            raise ValueError("Checkpoint callback is not allowed when checkpointing is disabled.")
        else:
            # if no checkpoint callback is defined, skip
            new_callback = NilCallback()

        # merge the checkpoint callback with other callbacks
        if callbacks is None:
            cbs = [new_callback]
        elif isinstance(callbacks, Callback):
            cbs = [callbacks, new_callback]
        else:
            cbs = callbacks + [new_callback]

        # set learning rate logger
        cbs.extend([LrLogger(), EpochNumLogger(), EpochTimeLogger()])

        # if gpumonitor is installed, enable callback
        try:
            gpu_cb = GpuStatsLogger()
        except ImportError:
            pass
        else:
            cbs.append(gpu_cb)

        # if psutil is installed, enable callback
        try:
            sys_cb = SystemStatsLogger()
        except ImportError:
            pass
        else:
            cbs.append(sys_cb)

        # setup the log mode
        self.log_every_n_steps = log_every_n_steps
        self.log_on_epoch = log_every_n_steps == 0
        self.log_on_step = log_every_n_steps > 0
        if self.log_on_epoch:
            log_every_n_steps = 10000000

        super().__init__(
            logger=logger is not None,
            enable_checkpointing=enable_checkpointing,
            callbacks=cbs,
            accelerator=accelerator,
            strategy=strategy,
            devices=devices,
            num_nodes=num_nodes,
            precision=precision,
            fast_dev_run=fast_dev_run,
            max_epochs=max_epochs,
            min_epochs=min_epochs,
            max_steps=max_steps,
            min_steps=min_steps,
            max_time=max_time,
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            limit_predict_batches=limit_predict_batches,
            overfit_batches=overfit_batches,
            val_check_interval=val_check_interval,
            check_val_every_n_epoch=check_val_every_n_epoch,
            num_sanity_val_steps=num_sanity_val_steps,
            log_every_n_steps=log_every_n_steps,
            enable_progress_bar=enable_progress_bar,
            enable_model_summary=enable_model_summary,
            accumulate_grad_batches=accumulate_grad_batches,
            gradient_clip_val=gradient_clip_val,
            gradient_clip_algorithm=gradient_clip_algorithm,
            deterministic=deterministic,
            benchmark=benchmark,
            inference_mode=inference_mode,
            use_distributed_sampler=use_distributed_sampler,
            profiler=profiler,
            detect_anomaly=detect_anomaly,
            barebones=barebones,
            plugins=plugins,
            sync_batchnorm=sync_batchnorm,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
            default_root_dir=default_root_dir,
        )

        # super().__init__(
        #     logger=logger is not None,
        #     enable_checkpointing=enable_checkpointing,
        #     callbacks=cbs,
        #     default_root_dir=default_root_dir,
        #     gradient_clip_val=gradient_clip_val,
        #     gradient_clip_algorithm=gradient_clip_algorithm,
        #     num_nodes=num_nodes,
        #     num_processes=num_processes,
        #     devices=devices,
        #     gpus=gpus,
        #     auto_select_gpus=auto_select_gpus,
        #     tpu_cores=tpu_cores,
        #     ipus=ipus,
        #     enable_progress_bar=enable_progress_bar,
        #     overfit_batches=overfit_batches,
        #     track_grad_norm=track_grad_norm,
        #     check_val_every_n_epoch=check_val_every_n_epoch,
        #     fast_dev_run=fast_dev_run,
        #     accumulate_grad_batches=accumulate_grad_batches,
        #     max_epochs=max_epochs,
        #     min_epochs=min_epochs,
        #     max_steps=max_steps,
        #     min_steps=min_steps,
        #     max_time=max_time,
        #     limit_train_batches=limit_train_batches,
        #     limit_val_batches=limit_val_batches,
        #     limit_test_batches=limit_test_batches,
        #     limit_predict_batches=limit_predict_batches,
        #     val_check_interval=val_check_interval,
        #     log_every_n_steps=log_every_n_steps,
        #     accelerator=accelerator,
        #     strategy=strategy,
        #     sync_batchnorm=sync_batchnorm,
        #     precision=precision,
        #     enable_model_summary=enable_model_summary,
        #     weights_save_path=weights_save_path,
        #     num_sanity_val_steps=num_sanity_val_steps,
        #     resume_from_checkpoint=resume_from_checkpoint,
        #     profiler=profiler,
        #     benchmark=benchmark,
        #     deterministic=deterministic,
        #     reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
        #     auto_lr_find=auto_lr_find,
        #     replace_sampler_ddp=replace_sampler_ddp,
        #     detect_anomaly=detect_anomaly,
        #     auto_scale_batch_size=auto_scale_batch_size,
        #     plugins=plugins,
        #     amp_backend=amp_backend,
        #     amp_level=amp_level,
        #     move_metrics_to_cpu=move_metrics_to_cpu,
        #     multiple_trainloader_mode=multiple_trainloader_mode
        # )
        # init loggers for training and validation
        self.has_no_logger = logger is None

        self.logger_train = TensorBoardLogger(save_dir=self.default_root_dir, name="logs",
            version=os.path.join(self.log_name, "train"), log_graph=True
        ) if self.has_no_logger is not None else None
        self.logger_val = TensorBoardLogger(save_dir=self.default_root_dir, name="logs",
            version=os.path.join(self.log_name, "val"), log_graph=False
        ) if self.has_no_logger is not None else None
        self._loggers = []

    @property
    def logger(self) -> Optional[Logger]:
        if self.has_no_logger:
            return None
        if self.training:
            return self.logger_train
        else:
            return self.logger_val



from typing import Union

import torch
from numpy import ndarray
from torch import Tensor


def iou_1d(proposal: Union[Tensor, ndarray], target: Union[Tensor, ndarray]) -> Tensor:
    """
    Calculate 1D IOU for N proposals with L labels.

    Args:
        proposal (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The predicted array with [M, 2]. First column is
            beginning, second column is end.
        target (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The label array with [N, 2]. First column is
            beginning, second column is end.

    Returns:
        :class:`~torch.Tensor`: The iou result with [M, N].
    """
    if type(proposal) is ndarray:
        proposal = torch.tensor(proposal)

    if type(target) is ndarray:
        target = torch.tensor(target)

    proposal_begin = proposal[:, 0].unsqueeze(0).T
    proposal_end = proposal[:, 1].unsqueeze(0).T
    target_begin = target[:, 0]
    target_end = target[:, 1]

    inner_begin = torch.maximum(proposal_begin, target_begin)
    inner_end = torch.minimum(proposal_end, target_end)
    outer_begin = torch.minimum(proposal_begin, target_begin)
    outer_end = torch.maximum(proposal_end, target_end)

    inter = torch.clamp(inner_end - inner_begin, min=0.)
    union = outer_end - outer_begin
    return inter / union


def iou_2d(proposal: Union[Tensor, ndarray], target: Union[Tensor, ndarray]) -> Tensor:
    """
    Calculate 2D IOU for M proposals with N targets.

    Args:
        proposal (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The proposals array with shape [M, 4]. The 4
            columns represents x1, y1, x2, y2.
        target (:class:`~torch.Tensor` | :class:`~numpy.ndarray`): The targets array with shape [N, 4]. The 4 columns
            represents x1, y1, x2, y2.

    Returns:
        :class:`~torch.Tensor`: The iou result with [M, N].
    """
    if type(proposal) is ndarray:
        proposal = torch.tensor(proposal)

    if type(target) is ndarray:
        target = torch.tensor(target)

    proposal_x1 = proposal[:, 0]
    proposal_y1 = proposal[:, 1]
    proposal_x2 = proposal[:, 2]
    proposal_y2 = proposal[:, 3]

    target_x1 = target[:, 0].unsqueeze(0).T
    target_y1 = target[:, 1].unsqueeze(0).T
    target_x2 = target[:, 2].unsqueeze(0).T
    target_y2 = target[:, 3].unsqueeze(0).T

    inner_x1 = torch.maximum(proposal_x1, target_x1)
    inner_y1 = torch.maximum(proposal_y1, target_y1)
    inner_x2 = torch.minimum(proposal_x2, target_x2)
    inner_y2 = torch.minimum(proposal_y2, target_y2)

    area_proposal = (proposal_x2 - proposal_x1) * (proposal_y2 - proposal_y1)
    area_target = (target_x2 - target_x1) * (target_y2 - target_y1)

    inter_x = torch.clamp(inner_x2 - inner_x1, min=0.)
    inter_y = torch.clamp(inner_y2 - inner_y1, min=0.)
    inter = inter_x * inter_y

    union = area_proposal + area_target - inter

    return inter / union



from typing import Any

from lightning.pytorch import Callback, Trainer, LightningModule
from lightning.pytorch.utilities.types import STEP_OUTPUT


class SystemStatsLogger(Callback):
    """Log system stats for each training epoch"""

    def __init__(self, on_epoch: bool = True, on_step: bool = False):
        try:
            import psutil
        except ImportError:
            raise ImportError("psutil is required to use SystemStatsLogger")
        self.psutil = psutil
        self.on_epoch = on_epoch
        self.on_step = on_step
        assert self.on_epoch or self.on_step, "on_epoch and on_step cannot be both False"

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if not self.on_epoch:
            return
        cpu_usage = self.psutil.cpu_percent()
        memory_usage = self.psutil.virtual_memory().percent
        logged_info = {
            "system/cpu_usage_epoch": cpu_usage,
            "system/memory_usage_epoch": memory_usage
        }
        pl_module.logger.log_metrics(logged_info, step=trainer.global_step)
        pl_module.log_dict(logged_info, logger=False, sync_dist=pl_module.distributed)

    def on_train_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        if not self.on_step:
            return
        cpu_usage = self.psutil.cpu_percent()
        memory_usage = self.psutil.virtual_memory().percent
        logged_info = {
            "system/cpu_usage_step": cpu_usage,
            "system/memory_usage_step": memory_usage
        }
        pl_module.logger.log_metrics(logged_info, step=trainer.global_step)
        pl_module.log_dict(logged_info, logger=False, sync_dist=pl_module.distributed)



import os.path

from tensorneko_util import io
from . import backend
from . import callback
from . import dataset
from . import debug
from . import evaluation
from . import layer
from . import module
from . import notebook
from . import optim
from . import preprocess
from . import util
from . import visualization
from .neko_model import NekoModel
from .neko_module import NekoModule
from .neko_trainer import NekoTrainer

__all__ = [
    "callback",
    "dataset",
    "backend",
    "evaluation",
    "io",
    "layer",
    "module",
    "notebook",
    "optim",
    "preprocess",
    "util",
    "visualization",
    "debug",
    "NekoModel",
    "NekoTrainer",
    "NekoModule"
]

__version__ = io.read.text(os.path.join(util.get_tensorneko_path(), "version.txt"))



from .round_robin_dataset import RoundRobinDataset
from .nested_dataset import NestedDataset
from .list_dataset import ListDataset
from . import sampler

__all__ = [
    "RoundRobinDataset",
    "NestedDataset",
    "ListDataset",
    "sampler"
]



from .log_graph import log_graph
from tensorneko_util.visualization import Colors, ContinuousColors, tensorboard
from . import watcher
from . import image_browser
from ..backend import VisualLib

__all__ = [
    "log_graph",
    "watcher",
    "image_browser",
    "tensorboard",
    "Colors",
    "ContinuousColors",
]

if VisualLib.matplotlib_available():
    from . import matplotlib
    from tensorneko_util.visualization import MultiPlots
    __all__.extend(["matplotlib", "MultiPlots"])

try:
    from tensorneko_util.visualization.seaborn import seaborn
except ImportError:
    pass
else:
    __all__.append("seaborn")



import random
from typing import Callable

import numpy as np
import torch
from torch import Generator


class Seed:
    """
    Seed manual setter to make program reproducible.

    Examples::

        # Set pytorch, numpy and python.random seed all to 42. And force PyTorch using the deterministic algorithms.
        Seed.set(42, use_deterministic_algorithms=True)

        # Apply seed to parallel workers of DataLoader
        DataLoader(
            train_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            worker_init_fn=Seed.get_loader_worker_init(),
            generator=Seed.get_torch_generator()
        )

        # See https://pytorch.org/docs/stable/notes/randomness.html

    References:
        Reproducibility — PyTorch 1.9.1 documentation. (2021). Retrieved 1 October 2021,
        from https://pytorch.org/docs/stable/notes/randomness.html
    """
    seed: int = None

    @classmethod
    def torch(cls, seed: int) -> None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)

    @classmethod
    def python(cls, seed: int) -> None:
        random.seed(seed)

    @classmethod
    def numpy(cls, seed: int) -> None:
        np.random.seed(seed)

    @classmethod
    def set(cls, seed: int, use_deterministic_algorithms: bool = False) -> None:
        cls.torch(seed)
        cls.python(seed)
        cls.numpy(seed)
        cls.seed = seed
        torch.use_deterministic_algorithms(use_deterministic_algorithms)

    @classmethod
    def _is_set(cls) -> bool:
        return cls.seed is not None

    @classmethod
    def get_loader_worker_init(cls) -> Callable[[int], None]:
        def seed_worker(worker_id):
            worker_seed = torch.initial_seed() % 2 ** 32
            np.random.seed(worker_seed)
            random.seed(worker_seed)

        if cls._is_set():
            return seed_worker
        else:
            return lambda x: None

    @classmethod
    def get_torch_generator(cls, device="cpu") -> Generator:
        g = torch.Generator(device)
        g.manual_seed(cls.seed)
        return g



from .gan import GAN
from .vqvae import VQVAE
from .wgan import WGAN
from .auto_encoder import AutoEncoder
from .binary_classifier import BinaryClassifier

__all__ = [
    "GAN",
    "VQVAE",
    "WGAN",
    "AutoEncoder",
    "BinaryClassifier",
]


from typing import Dict

from torch.nn import CrossEntropyLoss, L1Loss, MSELoss, BCELoss, BCEWithLogitsLoss, NLLLoss, \
    NLLLoss2d, PoissonNLLLoss, KLDivLoss, HingeEmbeddingLoss, MarginRankingLoss, \
    MultiLabelMarginLoss, MultiLabelSoftMarginLoss, MultiMarginLoss, SoftMarginLoss, \
    TripletMarginLoss, CosineEmbeddingLoss, HuberLoss, SmoothL1Loss
from torch.nn import LeakyReLU, GELU, ELU, ReLU, Mish, Module, ReLU6, PReLU, SELU, Sigmoid, \
    Tanh, Softplus, Softshrink, Softsign, Softmin, Softmax, LogSoftmax


class StringGetter:
    """Util class for retrieve module class from string"""

    _activation_mapping: Dict[str, Module] = {
        "LEAKYRELU": LeakyReLU,
        "GELU": GELU,
        "ELU": ELU,
        "RELU": ReLU,
        "RELU6": ReLU6,
        "PRELU": PReLU,
        "MISH": Mish,
        "SELU": SELU,
        "TANH": Tanh,
        "SIGMOID": Sigmoid,
        "SOFTPLUS": Softplus,
        "SOFTSHRINK": Softshrink,
        "SOFTSIGN": Softsign,
        "LOGSOFTMAX": LogSoftmax,
        "SOFTMIN": Softmin,
        "SOFTMAX": Softmax
    }

    _loss_mapping: Dict[str, Module] = {
        "CROSSENTROPYLOSS": CrossEntropyLoss,
        "L1LOSS": L1Loss,
        "MSELOSS": MSELoss,
        "BCELOSS": BCELoss,
        "BCEWITHLOGITSLOSS": BCEWithLogitsLoss,
        "NLLLOSS": NLLLoss,
        "NLLLOSS2D": NLLLoss2d,
        "POISSONNLLLOSS": PoissonNLLLoss,
        "KLDIVLOSS": KLDivLoss,
        "HINGEEMBEDDINGLOSS": HingeEmbeddingLoss,
        "MARGINRANKINGLOSS": MarginRankingLoss,
        "MULTIMARGINLOSS": MultiMarginLoss,
        "MULTILABELSOFTMARGINLOSS": MultiLabelSoftMarginLoss,
        "MULTILABELMARGINLOSS": MultiLabelMarginLoss,
        "SOFTMARGINLOSS": SoftMarginLoss,
        "TRIPLETMARGINLOSS": TripletMarginLoss,
        "COSINEEMBEDDINGLOSS": CosineEmbeddingLoss,
        "HUBERLOSS": HuberLoss,
        "SMOOTHL1LOSS": SmoothL1Loss
    }

    _mapping: Dict[str, Dict[str, Module]] = {
        "activation": _activation_mapping,
        "loss": _loss_mapping
    }

    def __init__(self, type_: str):
        self.type = type_
        self.dict = StringGetter._mapping[self.type]

    def __call__(self, name: str):
        return self._get(name)

    def _get(self, name: str):
        return self.dict[name.upper()]


get_activation = StringGetter("activation")
"""The function for getting activation function from its name"""


get_loss = StringGetter("loss")
"""The function for getting loss function from its name"""



from typing import List

from torch.utils.data.dataset import Dataset, T_co


class ListDataset(Dataset[T_co]):
    """
    A dataset wrapping a list of data.

    This class is a subclass of the PyTorch Dataset class and is used to wrap a list of data.
    It provides the necessary methods for a PyTorch Dataset, such as __getitem__ and __len__.

    Args:
        data (``List[T_co]``): The list of data that the dataset wraps.
    """

    def __init__(self, data: List[T_co]):
        """
        The constructor for the ListDataset class.

        Args:
            data (``List[T_co]``): The list of data that the dataset will wrap.
        """
        super().__init__()
        self.data = data

    def __getitem__(self, index: int) -> T_co:
        """
        Retrieves an item from the dataset at a specific index.

        Args:
            index (``int``): The index of the item to retrieve.

        Returns:
            The item at the specified index.
        """
        return self.data[index]

    def __len__(self):
        """
        Retrieves the length of the dataset.

        Returns:
            The length of the dataset (i.e., the number of items in the dataset).
        """
        return len(self.data)



from typing import Tuple

from torch import Tensor
from torch.nn import Conv2d
from torch.nn import functional as F

from ..neko_module import NekoModule
from ..layer.masked_conv2d import _VerticalStackConv2dA, _VerticalStackConv2dB, \
    _HorizontalStackConv2dA, _HorizontalStackConv2dB


class GatedConv(NekoModule):

    def __init__(self, mask_type: str, in_channels: int, kernel_size: int, residual: bool = True, **kwargs):
        super().__init__()
        assert mask_type in ["A", "B"]
        out_channels = in_channels * 2
        self.residual = residual
        self.Conv = {
            "vertical": _VerticalStackConv2dA if mask_type == "A" else _VerticalStackConv2dB,
            "horizontal": _HorizontalStackConv2dA if mask_type == "A" else _HorizontalStackConv2dB
        }

        self.conv_vertical = self.Conv["vertical"](in_channels, out_channels, kernel_size, **kwargs)
        self.conv_horizontal = self.Conv["horizontal"](in_channels, out_channels, kernel_size, **kwargs)
        self.vertical_to_horizontal = Conv2d(out_channels, out_channels, kernel_size=1)
        self.horizontal_proj = Conv2d(in_channels, in_channels, kernel_size=1)

    @staticmethod
    def activation(x: Tensor) -> Tensor:
        x, y = x.chunk(2, dim=1)
        return F.tanh(x) * F.sigmoid(y)

    def forward(self, x_v: Tensor, x_h: Tensor) -> Tuple[Tensor, Tensor]:
        # vertical stack
        v_feat = self.conv_vertical(x_v)
        v_out = self.activation(v_feat)

        # horizontal stack
        h_feat = self.conv_horizontal(x_h) + self.vertical_to_horizontal(v_feat)
        h_feat = self.activation(h_feat)
        h_out = self.horizontal_proj(h_feat)

        # residual connection
        if self.residual:
            h_out = h_out + x_h

        return h_out, v_out



import random
from typing import List, Optional

from torch.utils.data import Dataset
from torch.utils.data.dataset import T_co

from ..util import circular_pad


class RoundRobinDataset(Dataset[T_co]):
    """
    A dataset that combines multiple datasets into one by using a round-robin strategy.

    This class is a subclass of the PyTorch Dataset class and is used to combine multiple datasets.
    It provides the necessary methods for a PyTorch Dataset, such as `__getitem__` and `__len__`.

    Args:
        datasets (``List[Dataset]``): The input datasets.
        order (``List[int]``, optional): The order of the datasets indexes. For example,
            the datasets=[dataset0, dataset1], order=[1, 0, 1], means the data is sampled from
            dataset1, dataset0, dataset1, dataset1, dataset0, dataset1, ...
            Default: [0, 1, 2, ..., n] where `n` is the number of datasets.
        shuffle (``bool``, optional): Whether to shuffle the dataset indexes.
            Default: False.
        padding (``bool``, optional): True will pad the dataset to the longest length, False will cut the dataset to the
            shortest length. Default: True.

    """

    def __init__(self, datasets: List[Dataset[T_co]], order: Optional[List[int]] = None, shuffle: bool = False,
        padding: bool = True
    ):
        super().__init__()
        self.datasets = datasets
        self.order = order or list(range(len(datasets)))  # list of dataset indexes
        self.shuffle = shuffle
        self.padding = padding  # pad or cut

        self.dataset_lengths = [len(dataset) for dataset in datasets]

        # get the length for each sub dataset
        if padding:
            each_length = max(self.dataset_lengths)
        else:
            each_length = min(self.dataset_lengths)

        # total length for this dataset
        self.total_length = each_length * len(datasets)

        # generate the index sampler
        self.samplers = [list(range(length)) for length in self.dataset_lengths]

        if shuffle:
            for sampler in self.samplers:
                random.shuffle(sampler)

        if not padding:
            self.samplers = [sampler[:each_length] for sampler in self.samplers]
        else:
            for i in range(len(self.samplers)):
                self.samplers[i] = circular_pad(self.samplers[i], each_length)

    def __len__(self) -> int:
        """
        Retrieves the total length of the combined dataset.

        Returns:
            The total length of the combined dataset.
        """
        return self.total_length

    def __getitem__(self, index: int) -> T_co:
        """
        Retrieves an item from the combined dataset at a specific index.

        Args:
            index (int): The index of the item to retrieve.

        Returns:
            The item at the specified index.
        """
        dataset_index = self.order[index % len(self.order)]
        sample_index = self.samplers[dataset_index][index // len(self.order)]
        return self.datasets[dataset_index][sample_index]



from abc import abstractmethod, ABC
from typing import Optional, Union, Sequence, Dict, Tuple

import torch
from torch import Tensor
from torch.nn import Module
from torch.optim import Adam
from torch.optim.lr_scheduler import ReduceLROnPlateau

from ..neko_model import NekoModel
from ..layer import VectorQuantizer


class VQVAE(NekoModel, ABC):
    """
    Vector Quantized Variational Autoencoder (VQVAE). To use this model, you need to implement `build_encoder` and
    `build_decoder` methods.
    
    Args:
        latent_dim (``int``): Dimension of latent space.
        n_embeddings (``int``): Number of embeddings.
        learning_rate (``float``, optional): Learning rate. Default: 0.0003.
        *args: Other arguments for :class:`~tensorneko.neko_model.NekoModel`.
        **kwargs: Other keyword arguments for :class:`~tensorneko.neko_model.NekoModel`.

    Attributes:
        encoder (:class:`~torch.nn.Module`): Encoder.
        decoder (:class:`~torch.nn.Module`): Decoder.
        vector_quantizer (:class:`~tensorneko.layer.vector_quantizer.VectorQuantizer`): Vector quantizer.

    References:
        Oord, A., Vinyals, O., & Kavukcuoglu, K. (2017). Neural Discrete Representation Learning. Retrieved 18 May 2022,
        from https://arxiv.org/abs/1711.00937

    """

    def __init__(self, latent_dim: int, n_embeddings: int, beta: float = 0.25, learning_rate: float = 2e-4,
        name: str = "vqvae", *args, **kwargs
    ):
        super().__init__(name, *args, **kwargs)
        self.encoder = self.build_encoder()
        self.decoder = self.build_decoder()
        self.vector_quantizer = VectorQuantizer(n_embeddings, latent_dim, beta)

        self.learning_rate = learning_rate

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor]:
        z_e = self.encoder(x)
        z_q, embedding_loss = self.vector_quantizer(z_e)
        x_hat = self.decoder(z_q)
        return x_hat, embedding_loss

    @abstractmethod
    def build_encoder(self) -> Module:
        pass

    @abstractmethod
    def build_decoder(self) -> Module:
        pass

    @staticmethod
    def rec_loss_fn(pred: Tensor, target: Tensor) -> Tensor:
        return ((pred - target) ** 2).mean() / torch.var(target)

    def training_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        optimizer_idx: Optional[int] = None, hiddens: Optional[Tensor] = None
    ) -> Dict[str, Tensor]:
        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]
        x_hat, embedding_loss = self(x)
        rec_loss = self.rec_loss_fn(x_hat, x)
        loss = rec_loss + embedding_loss
        return {"loss": loss, "loss/r_loss": rec_loss, "loss/e_loss": embedding_loss}

    def validation_step(self, batch: Optional[Union[Tensor, Sequence[Tensor]]] = None, batch_idx: Optional[int] = None,
        dataloader_idx: Optional[int] = None
    ) -> Dict[str, Tensor]:
        if isinstance(batch, Tensor):
            x = batch
        else:
            x = batch[0]
        x_hat, embedding_loss = self(x)
        rec_loss = self.rec_loss_fn(x_hat, x)
        loss = rec_loss + embedding_loss
        return {"loss": loss, "loss/r_loss": rec_loss, "loss/e_loss": embedding_loss}

    def predict_step(self, batch: Tensor, batch_idx: int, dataloader_idx: Optional[int] = None) -> Tensor:
        return self(batch)[0]

    def configure_optimizers(self):
        optimizer = Adam(self.parameters(), lr=self.learning_rate)
        return {
            "optimizer": optimizer,
            "lr_scheduler": {
                "scheduler": ReduceLROnPlateau(optimizer, patience=5, factor=0.5, min_lr=1e-8),
                "monitor": "val_loss"
            }
        }



from abc import abstractmethod, ABC
from typing import Dict, Any, Callable, Tuple

from torch.nn import Module

from .util import F


class NekoModule(ABC, Module, F):
    """
    The module base of tensorneko. This NekoModule is inherited from :class:`~torch.nn.Module`
    and also support the ``tensorneko.util.fp`` pipe operation.

    Examples::

        # A three-layer-mlp with layer 16 -> 128 -> 1
        class ThreeLayerMLP(NekoModule):

            def __init__(self):
                super().__init__()
                self.linear0 = Linear(16, 128, build_activation=ReLU)
                self.linear1 = Linear(128, 1)

            def forward(self, x):
                f = self.linear0 >> self.linear1
                return f(x)

    """

    def __init__(self):
        super().__init__()
        self.f: Callable[[Tuple[Any, ...], Dict[str, Any]], Any] = self._call_impl

    @abstractmethod
    def forward(self, *args, **kwargs):
        ...

    @classmethod
    def __compose(cls, f, g):
        return F(lambda *args, **kwargs: f(g(*args, **kwargs)))

    def __ensure_callable(self, f):
        return F(*f) if isinstance(f, tuple) else f

    def __rshift__(self, g):
        """Overload << operator for F instances"""
        return self.__class__.__compose(self.__ensure_callable(g), self.f)

    def __lshift__(self, g):
        """Overload >> operator for F instances"""
        return self.__class__.__compose(self.f, self.__ensure_callable(g))



from abc import ABC, abstractmethod

import torch
from torch import Tensor, zeros
from torch.nn import Parameter, Dropout

from ..neko_module import NekoModule
from ..util import Shape, F, _


class PositionalEmbedding(NekoModule):
    """
    Trainable positional embedding for transformer/self-attention layer.

    The trainable positional embedding is used in Transformer (Vaswani, et al., 2017).

    Args:
        input_shape (:class:`~tensorneko.util.type.Shape`): The input shape of the sequence.
            Format: (Length of sequence, Embedding dimension).

        dropout_rate (``int``, optional): The dropout rate after the positional embedding. Default ``0``.

    Attributes:
        emb (:class`~torch.nn.Parameter`): The trainable positional embedding parameters.

        dropout (:class`~torch.nn.Dropout`): The PyTorch dropout module in the layer.

        trainable (``bool``): The trainable flag.

    References:
        Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
        Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

    """

    def __init__(self, input_shape: Shape, dropout_rate: float = 0.5, trainable: bool = True):
        super().__init__()
        self.input_shape = input_shape
        self.emb = Parameter(zeros(1, *input_shape), requires_grad=trainable)
        self.use_dropout = dropout_rate is not None and dropout_rate != 0.
        if self.use_dropout:
            self.dropout = Dropout(dropout_rate)

    def forward(self, x: Tensor) -> Tensor:
        x = x + self.emb
        if self.use_dropout:
            x = self.dropout(x)
        return x

    @property
    def trainable(self):
        return self.emb.requires_grad

    @trainable.setter
    def trainable(self, value: bool):
        self.emb.requires_grad = value


class AbstractNonTrainablePositionalEmbedding(PositionalEmbedding, ABC):
    """
    Abstract class for non-trainable positional embedding. Inherit this class to define your own positional embedding.

    The method `make_embedding` should be defined to generate the positional embedding as a Tensor with shape same to
        argument `input_shape`.

    Args:
        input_shape (:class:`~tensorneko.util.type.Shape`): The input shape of the sequence.
            Format: (Length of sequence, Embedding dimension).

        dropout_rate (``int``, optional): The dropout rate after the positional embedding. Default ``0``.

    Attributes:
        emb (:class`~torch.nn.Parameter`): The trainable positional embedding parameters.

        dropout (:class`~torch.nn.Dropout`): The PyTorch dropout module in the layer.

    """

    def __init__(self, input_shape: Shape, dropout_rate: float = 0.5):
        super().__init__(input_shape, dropout_rate, trainable=False)
        self.emb.data = self.make_embedding().unsqueeze(0)

    @abstractmethod
    def make_embedding(self) -> Tensor:
        pass


class SinCosPositionalEmbedding(AbstractNonTrainablePositionalEmbedding):
    """2D sine-cosine position embedding"""

    def make_embedding(self) -> Tensor:
        """
        Generate the positional embedding as a Tensor with shape same to argument `input_shape`.

        Pytorch version of
            https://github.com/jadore801120/attention-is-all-you-need-pytorch/blob/master/transformer/Models.py#L31

        Returns:
            :class:`~torch.Tensor`: The positional embedding.

        References:
            Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., ... & Polosukhin, I. (2017).
            Attention is all you need. In Advances in neural information processing systems (pp. 5998-6008).

        """

        n_position, d_hid = self.input_shape

        def get_position_angle_vec(position):
            return position / torch.tensor(10000).pow(
                2 * torch.div(torch.arange(d_hid), 2, rounding_mode='trunc') / d_hid)

        sinusoid_table = torch.stack([get_position_angle_vec(pos_i) for pos_i in range(n_position)], 0)
        sinusoid_table[:, 0::2] = torch.sin(sinusoid_table[:, 0::2])  # dim 2i
        sinusoid_table[:, 1::2] = torch.cos(sinusoid_table[:, 1::2])  # dim 2i+1

        return sinusoid_table.float()



from typing import List, Tuple, Union

import torch
from torch import Tensor

from ..neko_module import NekoModule
from ..util import F


class Stack(NekoModule):
    """
    The module version of :meth:`torch.stack` function family.

    Args:
        mode (``str``, optional): The mode of the pytorch stack type. Default original stack.
        dim (``int``, optional): The dimension of stack apply to. Cannot use in non-default mode. Default 0.

    Examples::

        dstack = Stack("d")
        x_stack = dstack([x1, x2])

    """

    def __init__(self, mode: str = "", dim: int = 0):
        super().__init__()
        # other mode cannot specify the dim
        assert not (mode != "" and dim != 0), "Other modes cannot specify the dim"
        if mode == "":
            self.stack_func = F(torch.stack, dim=dim)
        elif mode.lower() == "d":
            self.stack_func = torch.dstack
        elif mode.lower() == "v":
            self.stack_func = torch.vstack
        elif mode.lower() == "h":
            self.stack_func = torch.hstack
        elif mode.lower() == "column":
            self.stack_func = torch.column_stack
        elif mode.lower() == "row":
            self.stack_func = torch.row_stack
        else:
            raise ValueError("""Not a valid `mode` argument. It should be in ["", "d", "v", "h", "column", "row"].""")

    def forward(self, tensors: Union[List[Tensor], Tuple[Tensor, ...]]) -> Tensor:
        return self.stack_func(tensors)



from typing import Tuple, Union

import numpy as np
import torch
from PIL import Image
from einops import rearrange
from numpy import ndarray
from torch import Tensor, uint8, float32
from torch.nn import functional as func

from .enum import ResizeMethod, get_enum_value
from ..util import F, _


def resize_image(tensor: Tensor, size: Tuple[int, int], resize_method: Union[ResizeMethod, str] = ResizeMethod.BICUBIC
) -> Tensor:
    """
    Resizing an image to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Image tensor (C, H, W) with value range [0, 1].
        size ((``int``, ``int``)): Target size (H, W).
        resize_method (:class:`ResizeMethod` | ``str``, optional): Resize method. Default bicubic.

    Returns:
        :class:`~torch.Tensor`: The resized image.
    """
    h, w = size
    resize_method = _map_resize_method(resize_method)

    def resizer(x: ndarray) -> ndarray:
        x = Image.fromarray(x)
        x = x.resize((w, h), resample=resize_method.value)
        x = np.asarray(x).astype(np.uint8)
        return x

    f = F(rearrange, pattern="c h w -> h w c") \
        >> _ * 255 \
        >> (lambda x: x.to(uint8).numpy()) \
        >> resizer \
        >> torch.from_numpy \
        >> F(rearrange, pattern="h w c -> c h w") \
        >> _ / 255 \
        >> (lambda x: x.to(float32))
    return f(tensor)


def resize_video(tensor: Tensor, size: Tuple[int, int], resize_method: Union[ResizeMethod, str] = ResizeMethod.BICUBIC,
    fast=False
) -> Tensor:
    """
    Resizing a video to determined size.

    Args:
        tensor (:class:`~torch.Tensor`): Video tensor (T, C, H, W)
        size ((``int``, ``int``)): Target size (H, W)
        resize_method (:class:`ResizeMethod` | ``str``, optional): Resize method. Default bicubic.
        fast (bool, optional): If True, use fast mode (pytorch F.interpolate). Default False.

    Returns:
        :class:`~torch.Tensor`: The resized video.
    """
    if not fast:
        resize_method = _map_resize_method(resize_method)
        image_resizer = F(resize_image, size=size, resize_method=resize_method) \
                        >> F(rearrange, pattern="c h w -> 1 c h w")
        f = F(map, image_resizer) \
            >> list \
            >> torch.vstack
        return f(tensor)
    else:
        if resize_method == ResizeMethod.BICUBIC:
            resize_method = "bicubic"
        elif resize_method == ResizeMethod.BILINEAR:
            resize_method = "bilinear"
        elif resize_method == ResizeMethod.NEAREST:
            resize_method = "nearest"
        elif type(resize_method) == str:
            pass
        else:
            raise TypeError("Not matched resize method type.")

        return func.interpolate(tensor, size, mode=get_enum_value(resize_method, ResizeMethod), align_corners=False)


def _map_resize_method(resize_method: Union[ResizeMethod, str]) -> ResizeMethod:
    if resize_method == "bicubic":
        resize_method = ResizeMethod.BICUBIC
    elif resize_method == "bilinear":
        resize_method = ResizeMethod.BILINEAR
    elif resize_method == "nearest":
        resize_method = ResizeMethod.NEAREST
    elif resize_method == "lanczos":
        resize_method = ResizeMethod.LANCZOS
    elif resize_method == "box":
        resize_method = ResizeMethod.BOX
    elif type(resize_method) == ResizeMethod:
        pass
    else:
        raise TypeError("Not matched resize method type.")
    return resize_method
